@@ -21,23 +21,36 @@
  * OF THIS SOFTWARE.
  */
 
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
 #include <expat.h>
+#include <getopt.h>
 
 #include "wayland-util.h"
 
+enum side {
+	CLIENT,
+	SERVER,
+};
+
 static int
 usage(int ret)
 {
-	fprintf(stderr, "usage: ./scanner [client-header|server-header|code]\n");
+	fprintf(stderr, "usage: ./scanner [OPTION] [client-header|server-header|code]"
+		" [input_file output_file]\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Converts XML protocol descriptions supplied on "
-			"stdin to client headers,\n"
-			"server headers, or protocol marshalling code.\n");
+			"stdin or input file to client\n"
+			"headers, server headers, or protocol marshalling code.\n\n");
+	fprintf(stderr, "options:\n");
+	fprintf(stderr, "    -h,  --help                  display this help and exit.\n"
+		        "    -c,  --include-core-only     include the core version of the headers,\n"
+	                "                                 that is e.g. wayland-client-core.h instead\n"
+	                "                                 of wayland-client.h.\n");
 	exit(ret);
 }
 
@@ -61,6 +74,7 @@ struct protocol {
 	int null_run_length;
 	char *copyright;
 	struct description *description;
+	bool core_headers;
 };
 
 struct interface {
@@ -193,7 +207,7 @@ static const char *indent(int n)
 }
 
 static void
-desc_dump(char *desc, const char *fmt, ...) __attribute__((format(printf,2,3)));
+desc_dump(char *desc, const char *fmt, ...) WL_PRINTF(2, 3);
 
 static void
 desc_dump(char *desc, const char *fmt, ...)
@@ -283,7 +297,7 @@ warn(struct location *loc, const char *msg, ...)
 	va_end(ap);
 }
 
-static int
+static bool
 is_nullable_type(struct arg *arg)
 {
 	switch (arg->type) {
@@ -292,9 +306,9 @@ is_nullable_type(struct arg *arg)
 	case OBJECT:
 	case NEW_ID:
 	case ARRAY:
-		return 1;
+		return true;
 	default:
-		return 0;
+		return false;
 	}
 }
 
@@ -351,7 +365,7 @@ start_element(void *data, const char *element_name, const char **atts)
 		ctx->protocol->uppercase_name = uppercase_dup(name);
 		ctx->protocol->description = NULL;
 	} else if (strcmp(element_name, "copyright") == 0) {
-		
+
 	} else if (strcmp(element_name, "interface") == 0) {
 		if (name == NULL)
 			fail(&ctx->loc, "no interface name given");
@@ -399,11 +413,13 @@ start_element(void *data, const char *element_name, const char **atts)
 			message->destructor = 0;
 
 		if (since != NULL) {
+			int prev_errno = errno;
 			errno = 0;
 			version = strtol(since, &end, 0);
-			if (errno == EINVAL || end == since || *end != '\0')
+			if (errno != 0 || end == since || *end != '\0')
 				fail(&ctx->loc,
 				     "invalid integer (%s)\n", since);
+			errno = prev_errno;
 		} else {
 			version = 1;
 		}
@@ -543,6 +559,10 @@ end_element(void *data, const XML_Char *name)
 		   strcmp(name, "event") == 0) {
 		ctx->message = NULL;
 	} else if (strcmp(name, "enum") == 0) {
+		if (wl_list_empty(&ctx->enumeration->entry_list)) {
+			fail(&ctx->loc, "enumeration %s was empty",
+			     ctx->enumeration->name);
+		}
 		ctx->enumeration = NULL;
 	}
 }
@@ -846,16 +866,15 @@ emit_enumerations(struct interface *interface)
 }
 
 static void
-emit_structs(struct wl_list *message_list, struct interface *interface)
+emit_structs(struct wl_list *message_list, struct interface *interface, enum side side)
 {
 	struct message *m;
 	struct arg *a;
-	int is_interface, n;
+	int n;
 
 	if (wl_list_empty(message_list))
 		return;
 
-	is_interface = message_list == &interface->request_list;
 	if (interface->description) {
 		struct description *desc = interface->description;
 		printf("/**\n");
@@ -871,7 +890,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		printf(" */\n");
 	}
 	printf("struct %s_%s {\n", interface->name,
-	       is_interface ? "interface" : "listener");
+	       (side == SERVER) ? "interface" : "listener");
 
 	wl_list_for_each(m, message_list, link) {
 		struct description *mdesc = m->description;
@@ -880,8 +899,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		desc_dump(mdesc ? mdesc->summary : "(none)",
 			  "\t * %s - ", m->name);
 		wl_list_for_each(a, &m->arg_list, link) {
-
-			if (is_interface && a->type == NEW_ID &&
+			if (side == SERVER && a->type == NEW_ID &&
 			    a->interface_name == NULL)
 				printf("\t * @interface: name of the objects interface\n"
 				       "\t * @version: version of the objects interface\n");
@@ -901,7 +919,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		printf("\tvoid (*%s)(", m->name);
 
 		n = strlen(m->name) + 17;
-		if (is_interface) {
+		if (side == SERVER) {
 			printf("struct wl_client *client,\n"
 			       "%sstruct wl_resource *resource",
 			       indent(n));
@@ -914,14 +932,14 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		wl_list_for_each(a, &m->arg_list, link) {
 			printf(",\n%s", indent(n));
 
-			if (is_interface && a->type == OBJECT)
+			if (side == SERVER && a->type == OBJECT)
 				printf("struct wl_resource *");
-			else if (is_interface && a->type == NEW_ID && a->interface_name == NULL)
+			else if (side == SERVER && a->type == NEW_ID && a->interface_name == NULL)
 				printf("const char *interface, uint32_t version, uint32_t ");
-			else if (!is_interface && a->type == OBJECT && a->interface_name == NULL)
+			else if (side == CLIENT && a->type == OBJECT && a->interface_name == NULL)
 				printf("void *");
 
-			else if (!is_interface && a->type == NEW_ID)
+			else if (side == CLIENT && a->type == NEW_ID)
 				printf("struct %s *", a->interface_name);
 			else
 				emit_type(a);
@@ -934,7 +952,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 
 	printf("};\n\n");
 
-	if (!is_interface) {
+	if (side == CLIENT) {
 	    printf("static inline int\n"
 		   "%s_add_listener(struct %s *%s,\n"
 		   "%sconst struct %s_listener *listener, void *data)\n"
@@ -973,65 +991,6 @@ format_copyright(const char *copyright)
 	printf(" */\n\n");
 }
 
-static void
-emit_header(struct protocol *protocol, int server)
-{
-	struct interface *i;
-	const char *s = server ? "SERVER" : "CLIENT";
-
-	if (protocol->copyright)
-		format_copyright(protocol->copyright);
-
-	printf("#ifndef %s_%s_PROTOCOL_H\n"
-	       "#define %s_%s_PROTOCOL_H\n"
-	       "\n"
-	       "#ifdef  __cplusplus\n"
-	       "extern \"C\" {\n"
-	       "#endif\n"
-	       "\n"
-	       "#include <stdint.h>\n"
-	       "#include <stddef.h>\n"
-	       "#include \"%s\"\n\n"
-	       "struct wl_client;\n"
-	       "struct wl_resource;\n\n",
-	       protocol->uppercase_name, s,
-	       protocol->uppercase_name, s,
-	       server ? "wayland-util.h" : "wayland-client.h");
-
-	wl_list_for_each(i, &protocol->interface_list, link)
-		printf("struct %s;\n", i->name);
-	printf("\n");
-
-	wl_list_for_each(i, &protocol->interface_list, link) {
-		printf("extern const struct wl_interface "
-		       "%s_interface;\n",
-		       i->name);
-	}
-	printf("\n");
-
-	wl_list_for_each(i, &protocol->interface_list, link) {
-
-		emit_enumerations(i);
-
-		if (server) {
-			emit_structs(&i->request_list, i);
-			emit_opcodes(&i->event_list, i);
-			emit_opcode_versions(&i->event_list, i);
-			emit_event_wrappers(&i->event_list, i);
-		} else {
-			emit_structs(&i->event_list, i);
-			emit_opcodes(&i->request_list, i);
-			emit_stubs(&i->request_list, i);
-		}
-	}
-
-	printf("#ifdef  __cplusplus\n"
-	       "}\n"
-	       "#endif\n"
-	       "\n"
-	       "#endif\n");
-}
-
 static void
 emit_types_forward_declarations(struct protocol *protocol,
 				struct wl_list *message_list,
@@ -1067,6 +1026,102 @@ emit_types_forward_declarations(struct protocol *protocol,
 	}
 }
 
+static int
+cmp_names(const void *p1, const void *p2)
+{
+	const char * const *s1 = p1, * const *s2 = p2;
+
+	return strcmp(*s1, *s2);
+}
+
+static const char *
+get_include_name(bool core, enum side side)
+{
+	if (side == SERVER)
+		return core ? "wayland-server-core.h" : "wayland-server.h";
+	else
+		return core ? "wayland-client-core.h" : "wayland-client.h";
+}
+
+static void
+emit_header(struct protocol *protocol, enum side side)
+{
+	struct interface *i;
+	struct wl_array types;
+	const char *s = (side == SERVER) ? "SERVER" : "CLIENT";
+	char **p, *prev;
+
+	if (protocol->copyright)
+		format_copyright(protocol->copyright);
+
+	printf("#ifndef %s_%s_PROTOCOL_H\n"
+	       "#define %s_%s_PROTOCOL_H\n"
+	       "\n"
+	       "#ifdef  __cplusplus\n"
+	       "extern \"C\" {\n"
+	       "#endif\n"
+	       "\n"
+	       "#include <stdint.h>\n"
+	       "#include <stddef.h>\n"
+	       "#include \"%s\"\n\n"
+	       "struct wl_client;\n"
+	       "struct wl_resource;\n\n",
+	       protocol->uppercase_name, s,
+	       protocol->uppercase_name, s,
+	       get_include_name(protocol->core_headers, side));
+
+	wl_array_init(&types);
+	wl_list_for_each(i, &protocol->interface_list, link) {
+		emit_types_forward_declarations(protocol, &i->request_list, &types);
+		emit_types_forward_declarations(protocol, &i->event_list, &types);
+	}
+
+	wl_list_for_each(i, &protocol->interface_list, link) {
+		p = fail_on_null(wl_array_add(&types, sizeof *p));
+		*p = i->name;
+	}
+
+	qsort(types.data, types.size / sizeof *p, sizeof *p, cmp_names);
+	prev = NULL;
+	wl_array_for_each(p, &types) {
+		if (prev && strcmp(*p, prev) == 0)
+			continue;
+		printf("struct %s;\n", *p);
+		prev = *p;
+	}
+	wl_array_release(&types);
+	printf("\n");
+
+	wl_list_for_each(i, &protocol->interface_list, link) {
+		printf("extern const struct wl_interface "
+		       "%s_interface;\n", i->name);
+	}
+
+	printf("\n");
+
+	wl_list_for_each(i, &protocol->interface_list, link) {
+
+		emit_enumerations(i);
+
+		if (side == SERVER) {
+			emit_structs(&i->request_list, i, side);
+			emit_opcodes(&i->event_list, i);
+			emit_opcode_versions(&i->event_list, i);
+			emit_event_wrappers(&i->event_list, i);
+		} else {
+			emit_structs(&i->event_list, i, side);
+			emit_opcodes(&i->request_list, i);
+			emit_stubs(&i->request_list, i);
+		}
+	}
+
+	printf("#ifdef  __cplusplus\n"
+	       "}\n"
+	       "#endif\n"
+	       "\n"
+	       "#endif\n");
+}
+
 static void
 emit_null_run(struct protocol *protocol)
 {
@@ -1170,14 +1225,6 @@ emit_messages(struct wl_list *message_list,
 	printf("};\n\n");
 }
 
-static int
-cmp_names(const void *p1, const void *p2)
-{
-	const char * const *s1 = p1, * const *s2 = p2;
-
-	return strcmp(*s1, *s2);
-}
-
 static void
 emit_code(struct protocol *protocol)
 {
@@ -1246,31 +1293,80 @@ int main(int argc, char *argv[])
 {
 	struct parse_context ctx;
 	struct protocol protocol;
+	FILE *input = stdin;
 	int len;
 	void *buf;
+	bool help = false, core_headers = false;
+	bool fail = false;
+	int opt;
 	enum {
 		CLIENT_HEADER,
 		SERVER_HEADER,
 		CODE,
 	} mode;
 
-	if (argc != 2)
-		usage(EXIT_FAILURE);
-	else if (strcmp(argv[1], "help") == 0 || strcmp(argv[1], "--help") == 0)
+	static const struct option options[] = {
+		{ "help",              no_argument, NULL, 'h' },
+		{ "include-core-only", no_argument, NULL, 'c' },
+		{ 0,                   0,           NULL, 0 }
+	};
+
+	while (1) {
+		opt = getopt_long(argc, argv, "hc", options, NULL);
+
+		if (opt == -1)
+			break;
+
+		switch (opt) {
+		case 'h':
+			help = true;
+			break;
+		case 'c':
+			core_headers = true;
+			break;
+		default:
+			fail = true;
+			break;
+		}
+	}
+
+	argv += optind;
+	argc -= optind;
+
+	if (help)
 		usage(EXIT_SUCCESS);
-	else if (strcmp(argv[1], "client-header") == 0)
+	else if ((argc != 1 && argc != 3) || fail)
+		usage(EXIT_FAILURE);
+	else if (strcmp(argv[0], "help") == 0)
+		usage(EXIT_SUCCESS);
+	else if (strcmp(argv[0], "client-header") == 0)
 		mode = CLIENT_HEADER;
-	else if (strcmp(argv[1], "server-header") == 0)
+	else if (strcmp(argv[0], "server-header") == 0)
 		mode = SERVER_HEADER;
-	else if (strcmp(argv[1], "code") == 0)
+	else if (strcmp(argv[0], "code") == 0)
 		mode = CODE;
 	else
 		usage(EXIT_FAILURE);
 
+	if (argc == 3) {
+		input = fopen(argv[1], "r");
+		if (input == NULL) {
+			fprintf(stderr, "Could not open input file: %s\n",
+				strerror(errno));
+			exit(EXIT_FAILURE);
+		}
+		if (freopen(argv[2], "w", stdout) == NULL) {
+			fprintf(stderr, "Could not open output file: %s\n",
+				strerror(errno));
+			exit(EXIT_FAILURE);
+		}
+	}
+
 	wl_list_init(&protocol.interface_list);
 	protocol.type_index = 0;
 	protocol.null_run_length = 0;
 	protocol.copyright = NULL;
+	protocol.core_headers = core_headers;
 	memset(&ctx, 0, sizeof ctx);
 	ctx.protocol = &protocol;
 
@@ -1287,23 +1383,29 @@ int main(int argc, char *argv[])
 
 	do {
 		buf = XML_GetBuffer(ctx.parser, XML_BUFFER_SIZE);
-		len = fread(buf, 1, XML_BUFFER_SIZE, stdin);
+		len = fread(buf, 1, XML_BUFFER_SIZE, input);
 		if (len < 0) {
 			fprintf(stderr, "fread: %m\n");
 			exit(EXIT_FAILURE);
 		}
-		XML_ParseBuffer(ctx.parser, len, len == 0);
-
+		if (XML_ParseBuffer(ctx.parser, len, len == 0) == 0) {
+			fprintf(stderr,
+				"Error parsing XML at line %ld col %ld: %s\n",
+				XML_GetCurrentLineNumber(ctx.parser),
+				XML_GetCurrentColumnNumber(ctx.parser),
+				XML_ErrorString(XML_GetErrorCode(ctx.parser)));
+			exit(EXIT_FAILURE);
+		}
 	} while (len > 0);
 
 	XML_ParserFree(ctx.parser);
 
 	switch (mode) {
 		case CLIENT_HEADER:
-			emit_header(&protocol, 0);
+			emit_header(&protocol, CLIENT);
 			break;
 		case SERVER_HEADER:
-			emit_header(&protocol, 1);
+			emit_header(&protocol, SERVER);
 			break;
 		case CODE:
 			emit_code(&protocol);
