@@ -238,7 +238,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 	const struct wl_message *message;
 	uint32_t p[2];
 	uint32_t resource_flags;
-	int opcode, size;
+	int opcode, size, since;
 	int len;
 
 	if (mask & (WL_EVENT_ERROR | WL_EVENT_HANGUP)) {
@@ -260,7 +260,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 	len = 0;
 	if (mask & WL_EVENT_READABLE) {
 		len = wl_connection_read(connection);
-		if (len < 0 && errno != EAGAIN) {
+		if (len == 0 || (len < 0 && errno != EAGAIN)) {
 			wl_client_destroy(client);
 			return 1;
 		}
@@ -294,13 +294,14 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 		}
 
 		message = &object->interface->methods[opcode];
+		since = wl_message_get_since(message);
 		if (!(resource_flags & WL_MAP_ENTRY_LEGACY) &&
-		    resource->version > 0 &&
-		    resource->version < wl_message_get_since(message)) {
+		    resource->version > 0 && resource->version < since) {
 			wl_resource_post_error(client->display_resource,
 					       WL_DISPLAY_ERROR_INVALID_METHOD,
-					       "invalid method %d, object %s@%u",
-					       opcode,
+					       "invalid method %d (since %d < %d)"
+					       ", object %s@%u",
+					       opcode, resource->version, since,
 					       object->interface->name,
 					       object->id);
 			break;
@@ -358,7 +359,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
  * socket later - typically when the compositor has handled all
  * requests and goes back to block in the event loop.  This function
  * flushes all queued up events for a client immediately.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT void
@@ -371,7 +372,7 @@ wl_client_flush(struct wl_client *client)
  *
  * \param client The client object
  * \return The display object the client is associated with.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT struct wl_display *
@@ -402,7 +403,7 @@ bind_display(struct wl_client *client, struct wl_display *display);
  * WAYLAND_SOCKET environment variable on the client side.
  *
  * On failure this function sets errno accordingly and returns NULL.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT struct wl_client *
@@ -472,7 +473,7 @@ err_client:
  * then connects using socketpair(), this function will return the
  * credentials for the compositor.  The credentials for the socketpair
  * are set at creation time in the compositor.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT void
@@ -494,8 +495,8 @@ wl_client_get_credentials(struct wl_client *client,
  * \return The object or NULL if there is not object for the given ID
  *
  * This looks up an object in the client object name space by its
- * object ID.  
- * 
+ * object ID.
+ *
  * \memberof wl_client
  */
 WL_EXPORT struct wl_resource *
@@ -583,12 +584,12 @@ wl_resource_find_for_client(struct wl_list *list, struct wl_client *client)
 	if (client == NULL)
 		return NULL;
 
-        wl_list_for_each(resource, list, link) {
-                if (resource->client == client)
-                        return resource;
-        }
+	wl_list_for_each(resource, list, link) {
+		if (resource->client == client)
+			return resource;
+	}
 
-        return NULL;
+	return NULL;
 }
 
 WL_EXPORT struct wl_client *
@@ -670,7 +671,7 @@ wl_client_destroy(struct wl_client *client)
 	wl_map_for_each(&client->objects, destroy_resource, &serial);
 	wl_map_release(&client->objects);
 	wl_event_source_remove(client->source);
-	wl_connection_destroy(client->connection);
+	close(wl_connection_destroy(client->connection));
 	wl_list_remove(&client->link);
 	free(client);
 }
@@ -787,7 +788,6 @@ bind_display(struct wl_client *client, struct wl_display *display)
 
 /** Create Wayland display object.
  *
- * \param None
  * \return The Wayland display object. Null if failed to create
  *
  * This creates the wl_display object.
@@ -862,6 +862,20 @@ wl_socket_alloc(void)
 	return s;
 }
 
+/** Destroy Wayland display object.
+ *
+ * \param display The Wayland display object which should be destroyed.
+ * \return None.
+ *
+ * This function emits the wl_display destroy signal, releases
+ * all the sockets added to this display, free's all the globals associated
+ * with this display, free's memory of additional shared memory formats and
+ * destroy the display object.
+ *
+ * \sa wl_display_add_destroy_listener
+ *
+ * \memberof wl_display
+ */
 WL_EXPORT void
 wl_display_destroy(struct wl_display *display)
 {
@@ -938,7 +952,7 @@ wl_global_destroy(struct wl_global *global)
  *
  * This function returns the most recent serial number, but does not
  * increment it.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT uint32_t
@@ -953,7 +967,7 @@ wl_display_get_serial(struct wl_display *display)
  *
  * This function increments the display serial number and returns the
  * new value.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT uint32_t
@@ -1105,7 +1119,7 @@ wl_socket_init_for_display_name(struct wl_socket *s, const char *name)
 		 * "failed to add socket: Success" */
 		errno = ENAMETOOLONG;
 		return -1;
-	};
+	}
 
 	return 0;
 }
@@ -1126,7 +1140,7 @@ _wl_display_add_socket(struct wl_display *display, struct wl_socket *s)
 		return -1;
 	}
 
-	if (listen(s->fd, 1) < 0) {
+	if (listen(s->fd, 128) < 0) {
 		wl_log("listen() failed with error: %m\n");
 		return -1;
 	}
@@ -1181,6 +1195,30 @@ wl_display_add_socket_auto(struct wl_display *display)
 	return NULL;
 }
 
+/** Add a socket to Wayland display for the clients to connect.
+ *
+ * \param display Wayland display to which the socket should be added.
+ * \param name Name of the Unix socket.
+ * \return 0 if success. -1 if failed.
+ *
+ * This adds a Unix socket to Wayland display which can be used by clients to
+ * connect to Wayland display.
+ *
+ * If NULL is passed as name, then it would look for WAYLAND_DISPLAY env
+ * variable for the socket name. If WAYLAND_DISPLAY is not set, then default
+ * wayland-0 is used.
+ *
+ * The Unix socket will be created in the directory pointed to by environment
+ * variable XDG_RUNTIME_DIR. If XDG_RUNTIME_DIR is not set, then this function
+ * fails and returns -1.
+ *
+ * The length of socket path, i.e., the path set in XDG_RUNTIME_DIR and the
+ * socket name, must not exceed the maxium length of a Unix socket path.
+ * The function also fails if the user do not have write permission in the
+ * XDG_RUNTIME_DIR path or if the socket name is already in use.
+ *
+ * \memberof wl_display
+ */
 WL_EXPORT int
 wl_display_add_socket(struct wl_display *display, const char *name)
 {
@@ -1276,11 +1314,10 @@ wl_resource_create(struct wl_client *client,
 	resource->version = version;
 	resource->dispatcher = NULL;
 
-	if (wl_map_insert_at(&client->objects, 0, resource->object.id, resource) < 0) {
+	if (wl_map_insert_at(&client->objects, 0, id, resource) < 0) {
 		wl_resource_post_error(client->display_resource,
 				       WL_DISPLAY_ERROR_INVALID_OBJECT,
-				       "invalid new id %d",
-				       resource->object.id);
+				       "invalid new id %d", id);
 		free(resource);
 		return NULL;
 	}
@@ -1294,7 +1331,59 @@ wl_log_set_handler_server(wl_log_func_t handler)
 	wl_log_handler = handler;
 }
 
-/* Deprecated functions below. */
+/** Add support for a wl_shm pixel format
+ *
+ * \param display The display object
+ * \param format The wl_shm pixel format to advertise
+ * \return A pointer to the wl_shm format that was added to the list
+ * or NULL if adding it to the list failed.
+ *
+ * Add the specified wl_shm format to the list of formats the wl_shm
+ * object advertises when a client binds to it.  Adding a format to
+ * the list means that clients will know that the compositor supports
+ * this format and may use it for creating wl_shm buffers.  The
+ * compositor must be able to handle the pixel format when a client
+ * requests it.
+ *
+ * The compositor by default supports WL_SHM_FORMAT_ARGB8888 and
+ * WL_SHM_FORMAT_XRGB8888.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT uint32_t *
+wl_display_add_shm_format(struct wl_display *display, uint32_t format)
+{
+	uint32_t *p = NULL;
+
+	p = wl_array_add(&display->additional_shm_formats, sizeof *p);
+
+	if (p != NULL)
+		*p = format;
+	return p;
+}
+
+/**
+ * Get list of additional wl_shm pixel formats
+ *
+ * \param display The display object
+ *
+ * This function returns the list of addition wl_shm pixel formats
+ * that the compositor supports.  WL_SHM_FORMAT_ARGB8888 and
+ * WL_SHM_FORMAT_XRGB8888 are always supported and not included in the
+ * array, but all formats added through wl_display_add_shm_format()
+ * will be in the array.
+ *
+ * \sa wl_display_add_shm_format()
+ *
+ * \memberof wl_display
+ */
+struct wl_array *
+wl_display_get_additional_shm_formats(struct wl_display *display)
+{
+	return &display->additional_shm_formats;
+}
+
+/** \cond */ /* Deprecated functions below. */
 
 uint32_t
 wl_client_add_resource(struct wl_client *client,
@@ -1391,54 +1480,9 @@ wl_display_remove_global(struct wl_display *display, struct wl_global *global)
 	wl_global_destroy(global);
 }
 
-/** Add support for a wl_shm pixel format
- *
- * \param display The display object
- * \param format The wl_shm pixel format to advertise
- * \return A pointer to the wl_shm format that was added to the list
- * or NULL if adding it to the list failed.
- *
- * Add the specified wl_shm format to the list of formats the wl_shm
- * object advertises when a client binds to it.  Adding a format to
- * the list means that clients will know that the compositor supports
- * this format and may use it for creating wl_shm buffers.  The
- * compositor must be able to handle the pixel format when a client
- * requests it.
- *
- * The compositor by default supports WL_SHM_FORMAT_ARGB8888 and
- * WL_SHM_FORMAT_XRGB8888.
- *
- * \memberof wl_display
+/** \endcond */
+
+/* Functions at the end of this file are deprecated.  Instead of adding new
+ * code here, add it before the comment above that states:
+ * Deprecated functions below.
  */
-WL_EXPORT uint32_t *
-wl_display_add_shm_format(struct wl_display *display, uint32_t format)
-{
-	uint32_t *p = NULL;
-
-	p = wl_array_add(&display->additional_shm_formats, sizeof *p);
-
-	if (p != NULL)
-		*p = format;
-	return p;
-}
-
-/**
- * Get list of additional wl_shm pixel formats
- *
- * \param display The display object
- *
- * This function returns the list of addition wl_shm pixel formats
- * that the compositor supports.  WL_SHM_FORMAT_ARGB8888 and
- * WL_SHM_FORMAT_XRGB8888 are always supported and not included in the
- * array, but all formats added through wl_display_add_shm_format()
- * will be in the array.
- * 
- * \sa wl_display_add_shm_format()
- * 
- * \memberof wl_display
- */
-struct wl_array *
-wl_display_get_additional_shm_formats(struct wl_display *display)
-{
-	return &display->additional_shm_formats;
-}
