@@ -31,9 +31,11 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <poll.h>
 
 #include "wayland-private.h"
 #include "test-runner.h"
+#include "test-compositor.h"
 
 static const char message[] = "Hello, world";
 
@@ -57,6 +59,7 @@ TEST(connection_create)
 
 	connection = setup(s);
 	wl_connection_destroy(connection);
+	close(s[0]);
 	close(s[1]);
 }
 
@@ -74,6 +77,7 @@ TEST(connection_write)
 	assert(memcmp(message, buffer, sizeof message) == 0);
 
 	wl_connection_destroy(connection);
+	close(s[0]);
 	close(s[1]);
 }
 
@@ -92,6 +96,7 @@ TEST(connection_data)
 	wl_connection_consume(connection, sizeof message);
 
 	wl_connection_destroy(connection);
+	close(s[0]);
 	close(s[1]);
 }
 
@@ -117,6 +122,7 @@ TEST(connection_queue)
 	assert(memcmp(message, buffer + sizeof message, sizeof message) == 0);
 
 	wl_connection_destroy(connection);
+	close(s[0]);
 	close(s[1]);
 }
 
@@ -147,8 +153,8 @@ setup_marshal_data(struct marshal_data *data)
 static void
 release_marshal_data(struct marshal_data *data)
 {
-	wl_connection_destroy(data->read_connection);
-	wl_connection_destroy(data->write_connection);
+	close(wl_connection_destroy(data->read_connection));
+	close(wl_connection_destroy(data->write_connection));
 }
 
 static void
@@ -217,7 +223,7 @@ expected_fail_marshal(int expected_error, const char *format, ...)
 {
 	struct wl_closure *closure;
 	static const uint32_t opcode = 4444;
-	static const struct wl_interface test_interface = { 
+	static const struct wl_interface test_interface = {
 		.name = "test_object"
 	};
 	static struct wl_object sender = { 0 };
@@ -406,13 +412,13 @@ TEST(connection_demarshal)
 	msg[0] = 400200;
 	msg[1] = 12;
 	msg[2] = 0;
-	demarshal(&data, "?s", msg, (void *) validate_demarshal_s);	
+	demarshal(&data, "?s", msg, (void *) validate_demarshal_s);
 
 	release_marshal_data(&data);
 }
 
 static void
-marshal_demarshal(struct marshal_data *data, 
+marshal_demarshal(struct marshal_data *data,
 		  void (*func)(void), int size, const char *format, ...)
 {
 	struct wl_closure *closure;
@@ -495,7 +501,7 @@ TEST(connection_marshal_alot)
 	int i;
 
 	setup_marshal_data(&data);
-	
+
 	/* We iterate enough to make sure we wrap the circular buffers
 	 * for both regular data an fds. */
 
@@ -567,3 +573,80 @@ TEST(invoke_closure)
 {
 	marshal_helper("suu", suu_handler, "foo", 500, 404040);
 }
+
+static void
+leak_closure(void)
+{
+	struct wl_callback *cb;
+	struct pollfd pfd;
+	struct client *c = client_connect();
+
+	cb = wl_display_sync(c->wl_display);
+	assert(cb);
+	assert(wl_display_flush(c->wl_display) > 0);
+
+	/* we don't need it, it is referenced */
+	wl_callback_destroy(cb);
+
+	pfd.fd = wl_display_get_fd(c->wl_display);
+	pfd.events = POLLIN;
+
+	test_set_timeout(2);
+	assert(poll(&pfd, 1, -1) == 1);
+
+	/* read events, but do not dispatch them */
+	assert(wl_display_prepare_read(c->wl_display) == 0);
+	assert(wl_display_read_events(c->wl_display) == 0);
+
+	/*
+	 * now we have wl_callback.done and wl_display.delete_id queued;
+	 * if we now release the queue (in wl_display_disconnect())
+	 * we should not leak memory
+	 */
+
+	client_disconnect(c);
+}
+
+TEST(closure_leaks)
+{
+	struct display *d = display_create();
+
+	client_create(d, leak_closure);
+	display_run(d);
+
+	display_destroy(d);
+}
+
+static void
+leak_after_error(void)
+{
+	struct client *c = client_connect();
+
+	/* this should return -1, because we'll send error
+	 * from server. */
+	assert(stop_display(c, 1) == -1);
+	assert(wl_display_dispatch_pending(c->wl_display) == -1);
+	assert(wl_display_get_error(c->wl_display) == ENOMEM);
+
+	/* after we got error, we have display_resume event
+	 * in the queue. It should be freed in wl_display_disconnect().
+	 * Let's see! */
+
+	wl_proxy_destroy((struct wl_proxy *) c->tc);
+	wl_display_disconnect(c->wl_display);
+	free(c);
+}
+
+TEST(closure_leaks_after_error)
+{
+	struct display *d = display_create();
+	struct client_info *cl;
+
+	cl = client_create(d, leak_after_error);
+	display_run(d);
+
+	wl_client_post_no_memory(cl->wl_client);
+	display_resume(d);
+
+	display_destroy(d);
+}
